/*
 *        spFile.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>

#if !defined(MACOS9)
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include <sp/spBase.h>
#include <sp/spString.h>
#include <sp/spMemory.h>
#include <sp/spFile.h>
#include <sp/spOption.h>

#if defined(_WIN32) && !defined(__CYGWIN32__)
#include <mbctype.h>
#include <conio.h>
#endif

/*
 *        get file size
 */
spFileOffset spGetFileSize(const char *filename, long headsize)
{
    spFileOffset size;
    char *basename;

    /* get base name */
    basename = spGetBaseName(filename);

    if (streq(basename, "-") || streq(basename, "stdin")) {
#ifdef MACOS9
        return 0;
#else
#if defined(_WIN32) && !defined(__CYGWIN32__)
        struct _stat status;
#define fstat(h,buf) _fstat(h,buf)
#else
        struct stat status;
#endif
        
        /* stat file */
        if (fstat(0, &status)) {
            return 0;
        }
        size = status.st_size;
#endif
    } else {
        if (spGetFileInfo(filename, NULL, &size) == SP_FALSE) {
            return 0;
        }
    }

    /* get data size */
    size -= (spFileOffset)headsize;
    size = MAX(size, 0);

    return size;
}

long spGetTextFileSize(const char *filename)
{
    int ret;
    long size;
    double value;
    char line[SP_MAX_LINE];
    char string[SP_MAX_LINE];
    FILE *fp;

    if (NULL == (fp = spOpenFile(filename, "r"))) {
        return SP_FAILURE;
    }

    size = 0;
    while (spFGetLine(line, fp) != EOF) {
#if defined(HAVE_SSCANF_S)
        sscanf_s(line, "%s", string, (unsigned int)sizeof(string));
        ret = sscanf_s(string, "%lf", &value);
#else
        sscanf(line, "%s", string);
        ret = sscanf(string, "%lf", &value);
#endif
        if (ret == 1) {
            size++;
        }
    }

    spCloseFile(fp);

    return size;
}

/*
 *        skip header
 */
void spFSkipHeader(long headsize, FILE *fp)
{
    if (headsize >= 0)
        spSeekFile(fp, headsize, 0);
    return;
}

/*
 *        skip byte
 */
long spFSkipByte(long len, FILE *fp)
{
    long k;
    
    if (len > 0) {
        for (k = 0; k < len; k++) {
            if (spSeekFile(fp, 1L, 1) != 0) {
                break;
            }
        }
    } else {
        for (k = 0; k > len; k--) {
            if (spSeekFile(fp, -1L, 1) != 0) {
                break;
            }
        }
    }
    
    return k;
}

/*
 *        shift data
 */
void spShiftShort(short *data, long length, long shift)
{
    long k;
    long len;

    if (length <= 0 || shift == 0) return;

    if (data != NULL) {
        if (shift > 0) {
            for (k = length - 1; k >= 0; k--) {
                if (k >= shift) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0;
                }
            }
        } else {
            len = length + shift;
            for (k = 0; k < length; k++) {
                if (k < len) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0;
                }
            }
        }
    }

    return;
}

void spShiftLong(long *data, long length, long shift)
{
    long k;
    long len;

    if (length <= 0 || shift == 0) return;

    if (data != NULL) {
        if (shift > 0) {
            for (k = length - 1; k >= 0; k--) {
                if (k >= shift) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0;
                }
            }
        } else {
            len = length + shift;
            for (k = 0; k < length; k++) {
                if (k < len) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0;
                }
            }
        }
    }

    return;
}

void spShiftFloat(float *data, long length, long shift)
{
    long k;
    long len;

    if (length <= 0 || shift == 0) return;

    if (data != NULL) {
        if (shift > 0) {
            for (k = length - 1; k >= 0; k--) {
                if (k >= shift) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0.0;
                }
            }
        } else {
            len = length + shift;
            for (k = 0; k < length; k++) {
                if (k < len) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0.0;
                }
            }
        }
    }

    return;
}

void spShiftDouble(double *data, long length, long shift)
{
    long k;
    long len;

    if (length <= 0 || shift == 0) return;

    if (data != NULL) {
        if (shift > 0) {
            for (k = length - 1; k >= 0; k--) {
                if (k >= shift) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0.0;
                }
            }
        } else {
            len = length + shift;
            for (k = 0; k < length; k++) {
                if (k < len) {
                    data[k] = data[k - shift];
                } else {
                    data[k] = 0.0;
                }
            }
        }
    }

    return;
}

/*
 *        swap data
 */
void spSwapByte(void *data, long length, int samp_byte)
{
    long k;
    int i;
    int num_unit;
    int loop_count;
    char buf[16];
    char value;
    char *pos;

    pos = data;

    num_unit = samp_byte - 1;
    loop_count = samp_byte / 2;
    
    for (k = 0; k < length; k++) {
        memcpy(buf, pos, samp_byte);
        if (samp_byte == 2) {
            value = buf[0];
            buf[0] = buf[1];
            buf[1] = value;
        } else if (samp_byte == 3) {
            value = buf[0];
            buf[0] = buf[2];
            buf[2] = value;
        } else if (samp_byte == 4) {
            value = buf[0];
            buf[0] = buf[3];
            buf[3] = value;
            value = buf[1];
            buf[1] = buf[2];
            buf[2] = value;
        } else {
            for (i = 0; i < loop_count; i++) {
                value = buf[i];
                buf[i] = buf[num_unit - i];
                buf[num_unit - i] = value;
            }
        }
        memcpy(pos, buf, samp_byte);
        pos += samp_byte;
    }

    return;
}

void spSwapShort(short *data, long length)
{
    long k;

    for (k = 0; k < length; k++) {
        data[k] = ((data[k] << 8) & 0xff00) | ((data[k] >> 8) & 0xff);
    }

    return;
}

void spSwapLong(long *data, long length)
{
    long k;
    int i;
    int num_unit;
    char *pi, *po;
    long vi, vo;

    num_unit = sizeof(long) - 1;

    for (k = 0; k < length; k++) {
        vi = data[k];
        pi = (char *)&vi;
        po = (char *)&vo;
        for (i = 0; i <= num_unit; i++) {
            po[i] = pi[num_unit - i];
        }
        data[k] = vo;
    }
    
    return;
}

void spSwapLong32(spLong32 *data, long length)
{
    long k;
    int i;
    int num_unit;
    char *pi, *po;
    spLong32 vi, vo;

    num_unit = sizeof(spLong32) - 1;

    for (k = 0; k < length; k++) {
        vi = data[k];
        pi = (char *)&vi;
        po = (char *)&vo;
        for (i = 0; i <= num_unit; i++) {
            po[i] = pi[num_unit - i];
        }
        data[k] = vo;
    }
    
    return;
}

void spSwapULong32(spULong32 *data, long length)
{
    long k;
    int i;
    int num_unit;
    char *pi, *po;
    spULong32 vi, vo;

    num_unit = sizeof(spULong32) - 1;

    for (k = 0; k < length; k++) {
        vi = data[k];
        pi = (char *)&vi;
        po = (char *)&vo;
        for (i = 0; i <= num_unit; i++) {
            po[i] = pi[num_unit - i];
        }
        data[k] = vo;
    }
    
    return;
}

void spSwapFloat(float *data, long length)
{
    long k;
    int i;
    int num_unit;
    char *pi, *po;
    float vi, vo;

    num_unit = sizeof(float) - 1;

    for (k = 0; k < length; k++) {
        vi = data[k];
        pi = (char *)&vi;
        po = (char *)&vo;
        for (i = 0; i <= num_unit; i++) {
            po[i] = pi[num_unit - i];
        }
        data[k] = vo;
    }
    
    return;
}

void spSwapDouble(double *data, long length)
{
    long k;
    int i;
    int num_unit;
    char *pi, *po;
    double vi, vo;

    num_unit = sizeof(double) - 1;

    for (k = 0; k < length; k++) {
        vi = data[k];
        pi = (char *)&vi;
        po = (char *)&vo;
        for (i = 0; i <= num_unit; i++) {
            po[i] = pi[num_unit - i];
        }
        data[k] = vo;
    }
    
    return;
}

#ifdef SP_SUPPORT_64BIT_INT
void spSwapLong64(spLong64 *data, long length)
{
    long k;
    int i;
    int num_unit;
    char *pi, *po;
    spLong64 vi, vo;

    num_unit = sizeof(spLong64) - 1;

    for (k = 0; k < length; k++) {
        vi = data[k];
        pi = (char *)&vi;
        po = (char *)&vo;
        for (i = 0; i <= num_unit; i++) {
            po[i] = pi[num_unit - i];
        }
        data[k] = vo;
    }
    
    return;
}

void spSwapULong64(spULong64 *data, long length)
{
    long k;
    int i;
    int num_unit;
    char *pi, *po;
    spULong64 vi, vo;

    num_unit = sizeof(spULong64) - 1;

    for (k = 0; k < length; k++) {
        vi = data[k];
        pi = (char *)&vi;
        po = (char *)&vo;
        for (i = 0; i <= num_unit; i++) {
            po[i] = pi[num_unit - i];
        }
        data[k] = vo;
    }
    
    return;
}
#endif

/* for 64bit long represented by 4*char */
void spSwapLongC64(char *data_c64, long length)
{
    long k, l;
    long offset;
    char buf[8];

    for (k = 0, offset = 0; k < length; k++) {
        memcpy(buf, &data_c64[offset], 8 * sizeof(char));
        for (l = 0; l < 8; l++) {
            data_c64[offset + l] = buf[7 - l];
        }
        offset += 8;
    }

    return;
}

/*
 *        read data
 */
long spFReadShort(short *data, long length, int swap, FILE *fp)
{
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(short), length, fp)) <= 0) {
        return ndata;
    }

    if (swap) {
        /* byte swap */
        spSwapShort(data, ndata);
    }

    if (ndata < length) {
        long k;
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}

long spFReadLong(long *data, long length, int swap, FILE *fp)
{
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(long), length, fp)) <= 0) {
        return ndata;
    }

    if (swap) {
        /* byte swap */
        spSwapLong(data, ndata);
    }

    if (ndata < length) {
        long k;
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}

long spFReadLong32(long *data, long length, int swap, FILE *fp)
{
    long nread;
    long k;
    long ndata;
    spLong32 value;
    
    if (data == NULL) return 0;

    for (k = 0, ndata = 0; k < length; k++) {
        nread = (long)fread((char *)&value, sizeof(spLong32), 1, fp);
        if (nread <= 0) {
            spDebug(100, "spFReadLong32", "fread failed: %ld, length = %ld\n", nread, length);
            break;
        }

        if (swap) {
            /* byte swap */
            spSwapLong32(&value, 1);
        }

        data[k] = value;
        ++ndata;
    }

    if (ndata < length) {
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }
    
    spDebug(100, "spFReadLong32", "length = %ld, ndata = %ld\n", length, ndata);

    return ndata;
}

long spFReadULong32(unsigned long *data, long length, int swap, FILE *fp)
{
    long nread;
    long k;
    long ndata;
    spULong32 value;
    
    if (data == NULL) return 0;

    for (k = 0, ndata = 0; k < length; k++) {
        nread = (long)fread((char *)&value, sizeof(spULong32), 1, fp);
        if (nread <= 0) {
            spDebug(100, "spFReadULong32", "fread failed: %ld, length = %ld\n", nread, length);
            break;
        }

        if (swap) {
            /* byte swap */
            spSwapULong32(&value, 1);
        }

        data[k] = value;
        ++ndata;
    }

    if (ndata < length) {
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }
    
    spDebug(100, "spFReadULong32", "length = %ld, ndata = %ld\n", length, ndata);

    return ndata;
}

long spFReadLong32Direct(spLong32 *data, long length, int swap, FILE *fp)
{
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(spLong32), length, fp)) <= 0) {
        return ndata;
    }

    if (swap) {
        /* byte swap */
        spSwapLong32(data, ndata);
    }

    if (ndata < length) {
        long k;
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}

long spFReadULong32Direct(spULong32 *data, long length, int swap, FILE *fp)
{
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(spULong32), length, fp)) <= 0) {
        return ndata;
    }

    if (swap) {
        /* byte swap */
        spSwapULong32(data, ndata);
    }

    if (ndata < length) {
        long k;
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}

long spFReadLong24(long *data, long length, int swap, FILE *fp)
{
    char c;
    long k;
    long ndata;
    spLong32 value;
    char buf[4];
    
    if (data == NULL) return 0;

    for (k = 0, ndata = 0; k < length; k++) {
        if (fread(buf, 3, 1, fp) <= 0) {
            break;
        }

        if (swap) {
            c = buf[0]; buf[0] = buf[2]; buf[2] = c;
        }

        /* The divisor of 256 is required to keep sign bit */
#if BYTE_ORDER == LITTLE_ENDIAN
        value = (((spLong32)buf[2] << 24) & 0xff000000)
            | (((spLong32)buf[1] << 16) & 0xff0000) | (((spLong32)buf[0] << 8) & 0xff00);
#else
        value = (((spLong32)buf[0] << 24) & 0xff000000)
            | (((spLong32)buf[1] << 16) & 0xff0000) | (((spLong32)buf[2] << 8) & 0xff00);
#endif
        value /= 256;

        data[k] = value;
        ndata++;
    }

    if (ndata < length) {
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}

long spFReadULong24(unsigned long *data, long length, int swap, FILE *fp)
{
    char c;
    long k;
    long ndata;
    spULong32 value;
    char buf[4];
    
    if (data == NULL) return 0;

    for (k = 0, ndata = 0; k < length; k++) {
        if (fread(buf, 3, 1, fp) <= 0) {
            break;
        }

        if (swap) {
            c = buf[0]; buf[0] = buf[2]; buf[2] = c;
        }
        
        /* The divisor of 256 is required to keep sign bit */
#if BYTE_ORDER == LITTLE_ENDIAN
        value = (((spULong32)buf[2] << 24) & 0xff000000)
            | (((spULong32)buf[1] << 16) & 0xff0000) | (((spULong32)buf[0] << 8) & 0xff00);
#else
        value = (((spULong32)buf[0] << 24) & 0xff000000)
            | (((spULong32)buf[1] << 16) & 0xff0000) | (((spULong32)buf[2] << 8) & 0xff00);
#endif
        value /= 256;

        data[k] = value;
        ndata++;
    }

    if (ndata < length) {
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}

#ifdef SP_SUPPORT_64BIT_INT
long spFReadLong64(spLong64 *data, long length, int swap, FILE *fp)
{
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(spLong64), length, fp)) <= 0) {
        return ndata;
    }

    if (swap) {
        /* byte swap */
        spSwapLong64(data, ndata);
    }

    if (ndata < length) {
        long k;
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}

long spFReadULong64(spULong64 *data, long length, int swap, FILE *fp)
{
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(spULong64), length, fp)) <= 0) {
        return ndata;
    }

    if (swap) {
        /* byte swap */
        spSwapLong64((spLong64 *)data, ndata);
    }

    if (ndata < length) {
        long k;
        for (k = ndata; k < length; k++) {
            data[k] = 0;
        }
    }

    return ndata;
}
#endif

long spFReadFloatWeighted(float *data, long length, float weight, int swap, FILE *fp)
{
    long k;
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(float), length, fp)) <= 0) {
        return ndata;
    }
    if (swap) {
        /* byte swap */
        spSwapFloat(data, ndata);
    }
    if (weight != 1.0) {
        for (k = 0; k < ndata; k++) {
            data[k] *= weight;
        }
    }

    if (ndata < length) {
        for (k = ndata; k < length; k++) {
            data[k] = 0.0;
        }
    }

    return ndata;
}

long spFReadFloat(float *data, long length, int swap, FILE *fp)
{
    return spFReadFloatWeighted(data, length, 1.0, swap, fp);
}

long spFReadDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long k;
    long ndata = 0;
    
    if (data == NULL) return 0;
    
    /* read file */
    if ((ndata = (long)fread((char *)data, sizeof(double), length, fp)) <= 0) {
        spDebug(100, "spFReadDoubleWeighted", "failed: %ld, length = %ld\n", ndata, length);
        return ndata;
    }
    if (swap) {
        /* byte swap */
        spSwapDouble(data, ndata);
    }
    if (weight != 1.0) {
        for (k = 0; k < ndata; k++) {
            data[k] *= weight;
        }
    }

    if (ndata < length) {
        for (k = ndata; k < length; k++) {
            data[k] = 0.0;
        }
    }

    spDebug(100, "spFReadDoubleWeighted", "length = %ld, ndata = %ld\n", length, ndata);
    
    return ndata;
}

long spFReadDouble(double *data, long length, int swap, FILE *fp)
{
    return spFReadDoubleWeighted(data, length, 1.0, swap, fp);
}


long spFReadByte(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    unsigned char c;
    short value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        n = (long)fread((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0;
        } else {
            value = 256 * ((short)c - 128);
            data[k] = value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadSignedByte(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    signed char c;
    short value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        n = (long)fread((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0;
        } else {
            value = 256 * (short)c;
            data[k] = value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadULaw(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    unsigned char c;
    short value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        n = (long)fread((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0;
        } else {
            value = ulaw2linear(c);
            data[k] = value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadALaw(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    unsigned char c;
    short value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        n = (long)fread((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0;
        } else {
            value = alaw2linear(c);
            data[k] = value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadINT8(spINT8 *data, long length, FILE *fp)
{
    return (long)fread((char *)data, sizeof(spINT8), length, fp);
}

long spFReadINT16(spINT16 *data, long length, FILE *fp)
{
    return spFReadShort((short *)data, length, BYTE_ORDER == BIG_ENDIAN, fp);
}

long spFReadBYTE(spBYTE *data, long length, FILE *fp)
{
    return (long)fread((char *)data, sizeof(spBYTE), length, fp);
}

long spFReadWORD(spWORD *data, long length, FILE *fp)
{
    return spFReadShort((short *)data, length, BYTE_ORDER == BIG_ENDIAN, fp);
}

long spFReadLONG(spLONG *data, long length, FILE *fp)
{
#if defined(_WIN32)
    return spFReadLong32Direct(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#else
    return spFReadLong32(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#endif
}

long spFReadULONG(spULONG *data, long length, FILE *fp)
{
#if defined(_WIN32)
    return spFReadULong32Direct(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#else
    return spFReadULong32(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#endif
}

long spFReadShortToDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    short value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        n = (long)fread((char *)&value, sizeof(short), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0.0;
        } else {
            if (swap) {
                /* byte swap */
                spSwapShort(&value, 1);
            }
            
            data[k] = weight * (double)value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadShortToDouble(double *data, long length, int swap, FILE *fp)
{
    return spFReadShortToDoubleWeighted(data, length, 1.0, swap, fp);
}

long spFReadLongToDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    long value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        n = (long)fread((char *)&value, sizeof(long), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0.0;
        } else {
            if (swap) {
                /* byte swap */
                spSwapLong(&value, 1);
            }
            
            data[k] = weight * (double)value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadLongToDouble(double *data, long length, int swap, FILE *fp)
{
    return spFReadLongToDoubleWeighted(data, length, 1.0, swap, fp);
}

static long spFReadLong2432ToDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp, spBool long24_flag)
{
    long n = 0, ndata = 0;
    long k;
    long value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        if (long24_flag == SP_TRUE) {
            n = spFReadLong24(&value, 1, swap, fp);
        } else {
            n = spFReadLong32(&value, 1, swap, fp);
        }
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0.0;
        } else {
            data[k] = weight * (double)value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadLong32ToDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    return spFReadLong2432ToDoubleWeighted(data, length, weight, swap, fp, SP_FALSE);
}

long spFReadLong32ToDouble(double *data, long length, int swap, FILE *fp)
{
    return spFReadLong32ToDoubleWeighted(data, length, 1.0, swap, fp);
}

long spFReadLong24ToDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    return spFReadLong2432ToDoubleWeighted(data, length, weight, swap, fp, SP_TRUE);
}

long spFReadLong24ToDouble(double *data, long length, int swap, FILE *fp)
{
    return spFReadLong24ToDoubleWeighted(data, length, 1.0, swap, fp);
}

long spFReadFloatToDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    float value;
    
    if (data == NULL) return 0;

    for (k = 0; k < length; k++) {
        /* read file */
        n = (long)fread((char *)&value, sizeof(float), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            }
            data[k] = 0.0;
        } else {
            if (swap) {
                /* byte swap */
                spSwapFloat(&value, 1);
            }
            
            data[k] = weight * (double)value;
            ndata += n;
        }
    }

    return ndata;
}

long spFReadFloatToDouble(double *data, long length, int swap, FILE *fp)
{
    return spFReadFloatToDoubleWeighted(data, length, 1.0, swap, fp);
}

long spFReadBitToDoubleWeighted(double *data, long length, double weight, int samp_bit, int swap, FILE *fp)
{
    long nread;
    
    if (data == NULL || length <= 0) return 0;
    
    if (samp_bit >= 64) {
        nread = freaddoubleweighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 33) {
        nread = freadfloattodweighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 32) {
        nread = freadlong32todweighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 24) {
        nread = freadlong24todweighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 16){
        nread = freadshorttodweighted(data, length, weight, swap, fp);
    } else {
        nread = -1;
    }

    return nread;
}

long spFReadBitToDouble(double *data, long length, double max, int samp_bit, int swap, FILE *fp)
{
    double weight;
    
    if (data == NULL || length <= 0) return 0;

    weight = 1.0;
    
    if (samp_bit >= 33) {
        weight = max;
    } else if (samp_bit >= 32) {
        weight = max / 2147483648.0;
    } else if (samp_bit >= 24) {
        weight = max / 8388608.0;
    } else if (samp_bit >= 16){
        weight = max / 32768.0;
    } else {
        return -1;
    }

    return spFReadBitToDoubleWeighted(data, length, weight, samp_bit, swap, fp);
}

long spFReadFrameBit(double *data, long length, long shift, double max, int samp_bit, int swap, FILE *fp)
{
    long k;
    long offset;
    long nread;

    if (data == NULL) return 0;
    
    if (shift < 0) {
        offset = -shift;
        
        for (k = 0; k < offset; k++) {
            data[k] = 0.0;
        }
    } else {
        if (shift <= length) {
            offset = 0;
        } else {
            offset = shift;
        
            for (k = 0; k < offset; k++) {
                data[k] = data[k + shift];
            }
        }
    }
    
    nread = spFReadBitToDouble(data + offset, length - offset, max, samp_bit, swap, fp);

    if (nread < 0) {
        return nread;
    } else {
        return nread + offset;
    }
}


/*
 *        write data
 */
long spFWriteShort(short *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap) {
        long k;
        short value;

        for (k = 0; k < length; k++) {
            value = data[k];
            
            /* byte swap */
            spSwapShort(&value, 1);

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(short), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(short), length, fp);
    }

    return ndata;
}

long spFWriteLong(long *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap) {
        long k;
        long value;

        for (k = 0; k < length; k++) {
            value = data[k];
            
            /* byte swap */
            spSwapLong(&value, 1);

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(long), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(long), length, fp);
    }

    return ndata;
}

long spFWriteLong32(long *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    spLong32 value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (spLong32)data[k];

        if (swap) {
            /* byte swap */
            spSwapLong32(&value, 1);
        }

        /* write file */
        n = (long)fwrite((char *)&value, sizeof(spLong32), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteULong32(unsigned long *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    spULong32 value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (spULong32)data[k];

        if (swap) {
            /* byte swap */
            spSwapULong32(&value, 1);
        }

        /* write file */
        n = (long)fwrite((char *)&value, sizeof(spULong32), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteLong32Direct(spLong32 *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap) {
        long k;
        spLong32 value;

        for (k = 0; k < length; k++) {
            value = data[k];
            
            /* byte swap */
            spSwapLong32(&value, 1);

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(spLong32), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(spLong32), length, fp);
    }

    return ndata;
}

long spFWriteULong32Direct(spULong32 *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap) {
        long k;
        spULong32 value;

        for (k = 0; k < length; k++) {
            value = data[k];
            
            /* byte swap */
            spSwapULong32(&value, 1);

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(spULong32), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(spULong32), length, fp);
    }

    return ndata;
}

long spFWriteLong24(long *data, long length, int swap, FILE *fp)
{
    char c;
    long n = 0, ndata = 0;
    long k;
    spLong32 value;
    char buf[4];
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        /* The multiplier of 256 is required to keep sign bit */
        value = (spLong32)(data[k] * 256);

#if BYTE_ORDER == LITTLE_ENDIAN
        buf[2] = (char)((value & 0xff000000) >> 24);
        buf[1] = (char)((value & 0xff0000) >> 16);
        buf[0] = (char)((value & 0xff00) >> 8);
#else
        buf[0] = (char)((value & 0xff000000) >> 24);
        buf[1] = (char)((value & 0xff0000) >> 16);
        buf[2] = (char)((value & 0xff00) >> 8);
#endif

        if (swap) {
            c = buf[0]; buf[0] = buf[2]; buf[2] = c;
        }
        
        /* write file */
        n = (long)fwrite(buf, 3, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteULong24(unsigned long *data, long length, int swap, FILE *fp)
{
    char c;
    long n = 0, ndata = 0;
    long k;
    spULong32 value;
    char buf[4];
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        /* The multiplier of 256 is required to keep sign bit */
        value = (spULong32)(data[k] * 256);

#if BYTE_ORDER == LITTLE_ENDIAN
        buf[2] = (char)((value & 0xff000000) >> 24);
        buf[1] = (char)((value & 0xff0000) >> 16);
        buf[0] = (char)((value & 0xff00) >> 8);
#else
        buf[0] = (char)((value & 0xff000000) >> 24);
        buf[1] = (char)((value & 0xff0000) >> 16);
        buf[2] = (char)((value & 0xff00) >> 8);
#endif

        if (swap) {
            c = buf[0]; buf[0] = buf[2]; buf[2] = c;
        }
        
        /* write file */
        n = (long)fwrite(buf, 3, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

#ifdef SP_SUPPORT_64BIT_INT
long spFWriteLong64(spLong64 *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap) {
        long k;
        spLong64 value;

        for (k = 0; k < length; k++) {
            value = data[k];
            
            /* byte swap */
            spSwapLong64(&value, 1);

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(spLong64), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(spLong64), length, fp);
    }

    return ndata;
}

long spFWriteULong64(spULong64 *data, long length, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap) {
        long k;
        spULong64 value;

        for (k = 0; k < length; k++) {
            value = data[k];
            
            /* byte swap */
            spSwapULong64(&value, 1);

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(spULong64), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(spULong64), length, fp);
    }

    return ndata;
}
#endif

long spFWriteFloatWeighted(float *data, long length, float weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap) {
        long k;
        float value;

        for (k = 0; k < length; k++) {
            value = weight * data[k];
            
            /* byte swap */
            spSwapFloat(&value, 1);

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(float), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(float), length, fp);
    }

    return ndata;
}

long spFWriteFloat(float *data, long length, int swap, FILE *fp)
{
    return spFWriteFloatWeighted(data, length, 1.0, swap, fp);
}

long spFWriteDoubleWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    
    if (data == NULL) return 0;
    
    if (swap || weight != 1.0) {
        long k;
        double value;

        for (k = 0; k < length; k++) {
            value = weight * data[k];

            if (swap) {
                /* byte swap */
                spSwapDouble(&value, 1);
            }

            /* write file */
            n = (long)fwrite((char *)&value, sizeof(double), 1, fp);
            if (n <= 0) {
                if (k <= 0) {
                    return n;
                } else {
                    return ndata;
                }
            } else {
                ndata += n;
            }
        }
    } else {
        /* write file */
        ndata = (long)fwrite((char *)data, sizeof(double), length, fp);
    }

    return ndata;
}

long spFWriteDouble(double *data, long length, int swap, FILE *fp)
{
    return spFWriteDoubleWeighted(data, length, 1.0, swap, fp);
}

long spFWriteByte(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    unsigned char c;
    short value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (short)(data[k] / 256) + 128;
        c = (unsigned char)value;
        
        /* write file */
        n = (long)fwrite((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteSignedByte(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    signed char c;
    short value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (short)(data[k] / 256);
        c = (char)value;
        
        /* write file */
        n = (long)fwrite((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteULaw(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    unsigned char c;
    short value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = data[k];
        c = linear2ulaw(value);
        
        /* write file */
        n = (long)fwrite((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteALaw(short *data, long length, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    unsigned char c;
    short value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = data[k];
        c = linear2alaw(value);
        
        /* write file */
        n = (long)fwrite((char *)&c, 1, 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteINT8(spINT8 *data, long length, FILE *fp)
{
    return (long)fwrite((char *)data, sizeof(spINT8), length, fp);
}

long spFWriteINT16(spINT16 *data, long length, FILE *fp)
{
    return spFWriteShort((short *)data, length, BYTE_ORDER == BIG_ENDIAN, fp);
}

long spFWriteBYTE(spBYTE *data, long length, FILE *fp)
{
    return (long)fwrite((char *)data, sizeof(spBYTE), length, fp);
}

long spFWriteWORD(spWORD *data, long length, FILE *fp)
{
    return spFWriteShort((short *)data, length, BYTE_ORDER == BIG_ENDIAN, fp);
}

long spFWriteLONG(spLONG *data, long length, FILE *fp)
{
#if defined(_WIN32)
    return spFWriteLong32Direct(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#else
    return spFWriteLong32(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#endif
}

long spFWriteULONG(spULONG *data, long length, FILE *fp)
{
#if defined(_WIN32)
    return spFWriteULong32Direct(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#else
    return spFWriteULong32(data, length, BYTE_ORDER == BIG_ENDIAN, fp);
#endif
}

long spFWriteDoubleToShortWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    short value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (short)spRound(weight * data[k]);
        
        if (swap) {
            /* byte swap */
            spSwapShort(&value, 1);
        }

        /* write file */
        n = (long)fwrite((char *)&value, sizeof(short), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteDoubleToShort(double *data, long length, int swap, FILE *fp)
{
    return spFWriteDoubleToShortWeighted(data, length, 1.0, swap, fp);
}

long spFWriteDoubleToLongWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    long value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (long)spRound(weight * data[k]);
        
        if (swap) {
            /* byte swap */
            spSwapLong(&value, 1);
        }

        /* write file */
        n = (long)fwrite((char *)&value, sizeof(long), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteDoubleToLong(double *data, long length, int swap, FILE *fp)
{
    return spFWriteDoubleToLongWeighted(data, length, 1.0, swap, fp);
}

static long spFWriteDoubleToLong2432Weighted(double *data, long length, double weight, int swap, FILE *fp, spBool long24_flag)
{
    long n = 0, ndata = 0;
    long k;
    long value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (long)spRound(weight * data[k]);

        /* write file */
        if (long24_flag == SP_TRUE) {
            n = spFWriteLong24(&value, 1, swap, fp);
        } else {
            n = spFWriteLong32(&value, 1, swap, fp);
        }
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteDoubleToLong32Weighted(double *data, long length, double weight, int swap, FILE *fp)
{
    return spFWriteDoubleToLong2432Weighted(data, length, weight, swap, fp, SP_FALSE);
}

long spFWriteDoubleToLong32(double *data, long length, int swap, FILE *fp)
{
    return spFWriteDoubleToLong32Weighted(data, length, 1.0, swap, fp);
}

long spFWriteDoubleToLong24Weighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    long value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (long)spRound(weight * data[k]);

        /* write file */
        n = spFWriteLong24(&value, 1, swap, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteDoubleToLong24(double *data, long length, int swap, FILE *fp)
{
    return spFWriteDoubleToLong24Weighted(data, length, 1.0, swap, fp);
}

long spFWriteDoubleToFloatWeighted(double *data, long length, double weight, int swap, FILE *fp)
{
    long n = 0, ndata = 0;
    long k;
    float value;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        value = (float)(weight * data[k]);
        
        if (swap) {
            /* byte swap */
            spSwapFloat(&value, 1);
        }

        /* write file */
        n = (long)fwrite((char *)&value, sizeof(float), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}

long spFWriteDoubleToFloat(double *data, long length, int swap, FILE *fp)
{
    return spFWriteDoubleToFloatWeighted(data, length, 1.0, swap, fp);
}

long spFWriteDoubleToBitWeighted(double *data, long length, double weight, int samp_bit, int swap, FILE *fp)
{
    long nwrite;
    
    if (data == NULL || length <= 0) return 0;
    
    if (samp_bit >= 64) {
        nwrite = spFWriteDoubleWeighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 33) {
        nwrite = spFWriteDoubleToFloatWeighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 32) {
        nwrite = spFWriteDoubleToLong32Weighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 24) {
        nwrite = spFWriteDoubleToLong24Weighted(data, length, weight, swap, fp);
    } else if (samp_bit >= 16){
        nwrite = spFWriteDoubleToShortWeighted(data, length, weight, swap, fp);
    } else {
        nwrite = -1;
    }

    return nwrite;
}

/* max is converted into target maximum value
 * (e.g. the value is converted into 32767 in 16 bits case, 1.0 in floating cases. 1.0 means no conversion.) */
long spFWriteDoubleToBit(double *data, long length, double max, int samp_bit, int swap, FILE *fp)
{
    double weight;
    
    if (data == NULL || length <= 0) return 0;

    weight = 0.0;
    
    if (samp_bit >= 33) {
        if (max == 0.0) {
            weight = 1.0;
        } else {
            weight = 1.0 / max;
        }
    } else if (samp_bit >= 32) {
        weight = 2147483648.0 / max;
    } else if (samp_bit >= 24) {
        weight = 8388608.0 / max;
    } else if (samp_bit >= 16){
        weight = 32768.0 / max;
    } else {
        return -1;
    }

    return spFWriteDoubleToBitWeighted(data, length, weight, samp_bit, swap, fp);
}

static long spFWriteLongValueToShort(long *data, long length, int swap, FILE *fp, double value)
{
    long n = 0, ndata = 0;
    long k;
    spLong32 lvalue;
    short svalue;
    
    if (data == NULL) return 0;
    
    for (k = 0; k < length; k++) {
        lvalue = (spLong32)data[k];

        if (swap) {
            /* byte swap */
            spSwapLong32(&lvalue, 1);
        }

        svalue = (short)spRound((double)lvalue / value);

        /* write file */
        n = (long)fwrite((char *)&svalue, sizeof(short), 1, fp);
        if (n <= 0) {
            if (k <= 0) {
                return n;
            } else {
                return ndata;
            }
        } else {
            ndata += n;
        }
    }

    return ndata;
}


long spFWriteLong32ToShort(long *data, long length, int swap, FILE *fp)
{
    return spFWriteLongValueToShort(data, length, swap, fp, 65536.0);
}

long spFWriteLong24ToShort(long *data, long length, int swap, FILE *fp)
{
    return spFWriteLongValueToShort(data, length, swap, fp, 256.0);
}

long _spInterleaveData(char *ddata, long dlength, char *idata, long ilength, int num_channel, int samp_byte)
{
    int c;
    long k;
    long length, dlength2;
    long ntotal;
    char *uptr;

    length = MIN(dlength, ilength) / num_channel;
    dlength2 = dlength / (long)num_channel;
    ntotal = 0;
    
    for (k = 0; k < length; k++) {
        for (c = 0; c < num_channel; c++) {
            uptr = ddata + ((long)c * dlength2 + k) * (long)samp_byte;
            memcpy(idata, uptr, samp_byte);
            idata += samp_byte;
            ntotal++;
        }
    }
    
    return ntotal;
}

long _spDeinterleaveData(char *idata, long ilength, char *ddata, long dlength, int num_channel, int samp_byte)
{
    int c;
    long k;
    long length, dlength2;
    long ntotal;
    char *uptr;

    length = MIN(dlength, ilength) / (long)num_channel;
    dlength2 = dlength / (long)num_channel;
    ntotal = 0;
    
    for (k = 0; k < length; k++) {
        for (c = 0; c < num_channel; c++) {
            uptr = ddata + ((long)c * dlength2 + k) * (long)samp_byte;
            memcpy(uptr, idata, samp_byte);
            idata += samp_byte;
            ntotal++;
        }
    }
    
    return ntotal;
}


/*
 *        file I/O for text data
 */
int spGetNumRow(const char *filename)
{
    int nrow, ncol;
    FILE *fp;
    char buf[SP_MAX_LINE];

    if (NULL == (fp = spOpenFile(filename, "r"))) {
        return -1;
    }

    nrow = 0;
    while (1) {
        ncol = spFGetColumn(buf, 0, fp);
        if (ncol == EOF) {
            break;
        }
        nrow++;
    }

    spCloseFile(fp);

    return nrow;
}

int spGetNumColumn(const char *filename)
{
    int ncol;
    FILE *fp;
    char buf[SP_MAX_LINE];

    if (NULL == (fp = spOpenFile(filename, "r"))) {
        return -1;
    }

    ncol = spFGetColumn(buf, 0, fp);

    spCloseFile(fp);

    return ncol;
}

int spGetCSVNumRow(const char *filename)
{
    int nrow, ncol;
    FILE *fp;
    char buf[SP_MAX_LINE];

    if (NULL == (fp = spOpenFile(filename, "r"))) {
        return -1;
    }

    nrow = 0;
    while (1) {
        ncol = spFGetCSVNColumn(buf, sizeof(buf), 0, fp);
        if (ncol == EOF) {
            break;
        }
        nrow++;
    }

    spCloseFile(fp);

    return nrow;
}

int spGetCSVNumColumn(const char *filename)
{
    int ncol;
    FILE *fp;
    char buf[SP_MAX_LINE];

    if (NULL == (fp = spOpenFile(filename, "r"))) {
        return -1;
    }

    ncol = spFGetCSVNColumn(buf, sizeof(buf), 0, fp);

    spCloseFile(fp);

    return ncol;
}

#if defined(_WIN32) && !defined(__CYGWIN32__)
static int spfgetcwin(FILE *fp)
{
    if (fp != NULL) {
        return fgetc(fp);
    } else {
        return _getch();
    }
}

static int spungetcwin(int c, FILE *fp)
{
    if (fp != NULL) {
        return ungetc(c, fp);
    } else {
        return _ungetch(c);
    }
}

#define fgetc spfgetcwin
#define ungetc spungetcwin
#endif

static spBool isDelim(int c, const char *delim)
{
    int i;

    if (delim == NULL) {
        if (c == ' ' || c == '\t') {
            return SP_TRUE;
        } else {
            return SP_FALSE;
        }
    }

    for (i = 0; delim[i] != NUL; i++) {
        if ((char)c == delim[i]) {
            return SP_TRUE;
        }
    }

    return SP_FALSE;
}

/*
 *        get column from text file
 */
static int _spGetNColumn(char *buf, int size, int col, int *offset, const char *delim, const char *line, FILE *fp)
{
    int i;
    int end;
    int ncol;
    int nchar;
    int c, prev_c;
    int quote;
    spBool use_line;

    if (fp == NULL && line == NULL) return EOF;

    if (line != NULL) {
        use_line = SP_TRUE;
    } else {
        use_line = SP_FALSE;
    }

    i = 0;
    buf[0] = NUL;
    c = NUL;
    prev_c = NUL;
    quote = NUL;
    end = 0;
    ncol = 0;
    nchar = 0;
    while (1) {
        if (use_line == SP_FALSE) {
            c = fgetc(fp);
        } else {
            c = *line;
            line++;
        }
        i++;
        
        if (c == EOF || (use_line == SP_TRUE && c == NUL)) {
            c = EOF;
            break;
        } else if (quote == NUL && c == '#' && prev_c != '\\') {
            while (1) {
                if (use_line == SP_FALSE) {
                    c = fgetc(fp);
                } else {
                    c = *line;
                    line++;
                }
                i++;
                
                if (c == EOF || (use_line == SP_TRUE && c == NUL)) {
                    c = EOF;
                    end = 1;
                    break;
                } else if (c == '\n' || c == '\r') {
                    if (c == '\r') {
                        if (use_line == SP_FALSE) {
                            if ((c = fgetc(fp)) != '\n') { /* for Mac */
                                ungetc(c, fp);
                                i--;
                            }
                        } else {
                            if ((c = *(line++)) != '\n') {
                                line--;
                                i--;
                            }
                        }
                        i++;
                    }
                    if (ncol > 0) {
                        end = 1;
                    }
                    break;
                }
            }
            
            if (end == 1) {
                c = EOF;
                break;
            }
        } else if (quote == NUL && (c == '\n' || c == '\r')) {
            if (c == '\r') {
                if (use_line == SP_FALSE) {
                    if ((c = fgetc(fp)) != '\n') { /* for Mac */
                        ungetc(c, fp);
                        i--;
                    }
                } else {
                    if ((c = *(line++)) != '\n') {
                        line--;
                        i--;
                    }
                }
                i++;
            }
            if (prev_c == '\\' || (prev_c == NUL && nchar <= 0)/**/) {
            } else if (ncol != 0 || nchar != 0) {
                c = EOF;
                break;
            }
        } else if (quote == NUL && isDelim(c, delim) == SP_TRUE && prev_c != '\\') {
            if (nchar > 0) {
                if (nchar >= size) {
                    if (nchar == size) {
                        buf[nchar - 1] = NUL;
                    }
                } else {
                    if (ncol == col) {
                        buf[nchar] = NUL;
                    }
                }
                ncol++;
            }
            nchar = 0;
        } else if (c == '\\' && prev_c != '\\'
                   && spIsMBTailCandidate(prev_c, c) == SP_FALSE) {
        } else {
#if 0
            if (c == '\t') {
                c = ' ';
            }
#endif
            
            /*spDebug(100, "_spGetNColumn", "prev_c = %c, c = %c, quote = %c\n", prev_c, c, quote);*/
            
            if (prev_c != '\\'
                && spIsMBTailCandidate(prev_c, c) == SP_FALSE
                && ((quote != NUL && c == quote)
                    || (quote == NUL && (c == '\'' || c == '\"')))) {
                if (quote != NUL && c == quote) {
                    spDebug(60, "_spGetNColumn", "quotation %c finished\n", quote);
            
                    if (nchar <= 0) { /* no characters in this quoation */
                        if (ncol == col) {
                            buf[0] = NUL;
                        }
                        nchar = 1;
                    }
                    if (ncol == col && offset != NULL) {
                        if (use_line == SP_TRUE && *line == NUL) {
                            *offset = i;
                        } else {
                            *offset = i + 1;
                        }
                    }
                    quote = NUL;
                } else {
                    quote = c;
                    spDebug(60, "_spGetNColumn", "quotation %c started\n", quote);
                }
            } else {
                if (nchar >= size) {
                    if (nchar == size) {
                        buf[nchar - 1] = NUL;
                    }
                } else {
                    if (ncol == col) {
                        if (prev_c == '\\' && c == 'n') {
                            buf[nchar] = '\n';
                        } else {
                            buf[nchar] = c;
                        }
                        if (offset != NULL) {
                            spDebug(100, "_spGetNColumn", "col = %d, nchar = %d, i = %d, c = %c\n", col, nchar, i, c);
                            if (use_line == SP_TRUE && *line == NUL) {
                                *offset = i;
                            } else {
                                *offset = i + 1;
                            }
                        }
                    }
                }
                nchar++;

                if (c == '\\') {
                    c = NUL;
                }
            }
        }

        if (spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
            prev_c = 0;
        } else {
            prev_c = c;
        }
    }

    if (nchar > 0) {
        if (ncol == col) {
            nchar = MIN(nchar, size - 1);
            for (i = nchar - 1; i >= 0; i--) {
                if (isDelim(buf[i], delim) == SP_TRUE) {
                    --nchar;
                } else {
                    break;
                }
            }

            if ((nchar >= 1 && buf[nchar - 1] == '\\') ||
                nchar == 0) {
                ++nchar;
            }
            
            buf[nchar] = NUL;
        }
        ncol++;
    }

    if (ncol <= col && c == EOF) {
        ncol = EOF;
    }

    return ncol;
}

int spFGetNColumn(char *buf, int size, int col, FILE *fp)
{
    return _spGetNColumn(buf, size, col, NULL, NULL, NULL, fp);
}

int spFGetColumn(char *buf, int col, FILE *fp)
{
    return spFGetNColumn(buf, SP_MAX_LINE, col, fp);
}

int spSGetNColumn(char *buf, int size, int col, const char *line)
{
    return _spGetNColumn(buf, size, col, NULL, NULL, line, NULL);
}

int spSGetColumn(char *buf, int col, const char *line)
{
    return spSGetNColumn(buf, SP_MAX_LINE, col, line);
}
    
char *spSGetNextNColumn(char *buf, int size, const char *line)
{
    int offset = 0;
    int ncol = -1;

    if (line != NULL) {
        if ((ncol = _spGetNColumn(buf, size, 0, &offset, NULL, line, NULL)) >= 1
            && offset <= (int)strlen(line)) {
            /*spDebug(80, "spSGetNextNColumn", "offset = %d, line = %s\n", offset, line);*/
            return (char *)(line + offset);
        }
    }

    return NULL;
}

int spFGetCSVNColumn(char *buf, int size, int col, FILE *fp)
{
    return _spGetNColumn(buf, size, col, NULL, ",", NULL, fp);
}

int spSGetCSVNColumn(char *buf, int size, int col, const char *line)
{
    return _spGetNColumn(buf, size, col, NULL, ",", line, NULL);
}

char *spSGetNextCSVNColumn(char *buf, int size, const char *line)
{
    int offset = 0;
    int ncol = -1;
    int len = 0;

    if (line != NULL) {
        if ((ncol = _spGetNColumn(buf, size, 0, &offset, ",", line, NULL)) >= 1
            && offset <= (len = (int)strlen(line))) {
            spDebug(80, "spSGetNextCSVNColumn", "ncol = %d, offset = %d / %d, line = %s\n", ncol, offset, len, line);
            return (char *)(line + offset);
        }
        spDebug(80, "spSGetNextCSVNColumn", "failed: ncol = %d, offset = %d / %d, line = %s\n", ncol, offset, len, line);
    }
    
    return NULL;
}

int spFGetNLine(char *buf, int size, FILE *fp)
{
    int c, prev_c;
    int end;
    int ncol;
    int nchar;
    int pos;
    int quote;

    buf[0] = NUL;
    c = NUL;
    prev_c = NUL;
    quote = NUL;
    end = 0;
    ncol = 0;
    nchar = 0;
    pos = 0;
    while (1) {
        if (pos >= size) {
            pos--;
            break;
        }
        
        c = fgetc(fp);
        if (c == EOF) {
            break;
        } else if (quote == NUL && c == '#' && prev_c != '\\') {
            while (1) {
                c = fgetc(fp);
                if (c == EOF) {
                    end = 1;
                    break;
                } else if (c == '\n' || c == '\r') {
                    if (c == '\r' && (c = fgetc(fp)) != '\n') { /* for Mac */
                        ungetc(c, fp);
                    }
                    if (ncol > 0) {
                        end = 1;
                    }
                    break;
                }
            }
            
            if (end == 1) {
                break;
            }
        } else if (quote == NUL && c == '\n' || c == '\r') {
            if (c == '\r' && (c = fgetc(fp)) != '\n') { /* for Mac */
                ungetc(c, fp);
            }
            if (prev_c == '\\') {
                pos--;
                nchar--;
            } else if (ncol != 0 || nchar != 0) {
                break;
            }
        } else if (quote == NUL && c == ' ' || c == '\t') {
            if (nchar > 0) {
                ncol++;
            }
            buf[pos] = c;
            pos++;
            nchar = 0;
        } else {
            if (prev_c != '\\'
                && spIsMBTailCandidate(prev_c, c) == SP_FALSE
                && ((quote != NUL && c == quote)
                    || (quote == NUL && (c == '\'' || c == '\"')))) {
                if (quote != NUL && c == quote) {
                    spDebug(60, "spFGetNLine", "quotation %c finished\n", quote);
                    quote = NUL;
                } else {
                    quote = c;
                    spDebug(60, "spFetNLine", "quotation %c started\n", quote);
                }
            }
            
            buf[pos] = c;
            pos++;
            nchar++;

            if (spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
                c = NUL;
            }
        }

        prev_c = c;
    }

    buf[pos] = NUL;

    if (pos <= 0 && c == EOF) {
        pos = EOF;
    }

    return pos;
}

int spFGetLine(char *buf, FILE *fp)
{
    return spFGetNLine(buf, SP_MAX_LINE, fp);
}

int spGetNLine(char *buf, int size)
{
    return spFGetNLine(buf, size, spgetstdin());
}

int spGetLine(char *buf)
{
    return spFGetLine(buf, spgetstdin());
}

int spFPutLine(const char *input, FILE *fp)
{
    int i, j;
    int prev_c, c;

    if (input == NULL) return -1;
    
    prev_c = 0;
    for (i = 0, j = 0;; i++) {
        c = input[i];
        if (c == NUL) {
            break;
        } else if (c == '\n') {
            putc('\\', fp);
            putc('n', fp);
            j += 2;
        } else if (c == '#'
                   || (c == '\\'
                       && spIsMBTailCandidate(prev_c, c) == SP_FALSE)
                   || c == '\'' || c == '\"') {
            putc('\\', fp);
            putc(c, fp);
            j += 2;
        } else {
            putc(c, fp);
            if (spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
                c = NUL;
            }
            j++;
        }
        prev_c = c;
    }

    return j;
}

int spSPutLine(const char *input, char *line, int size)
{
    int i, j;
    int prev_c, c;

    if (input == NULL || line == NULL) return -1;
    
    prev_c = 0;
    for (i = 0, j = 0;; i++) {
        c = input[i];
        if (c == NUL) {
            break;
        } else if (c == '\n') {
            line[j] = '\\';
            line[j+1] = 'n';
            j += 2;
        } else if (c == '#'
                   || (c == '\\'
                       && spIsMBTailCandidate(prev_c, c) == SP_FALSE)
                   || c == '\'' || c == '\"') {
            if (j + 2 >= size) {
                break;
            }
            line[j] = '\\';
            line[j+1] = c;
            j += 2;
        } else {
            if (j + 1 >= size) {
                break;
            }
            line[j] = c;
            if (spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
                c = NUL;
            }
            j++;
        }
        prev_c = c;
    }

    line[j] = NUL;

    return j;
}

static int _spPutString(const char *input, char *line, int size, FILE *fp)
{
    int i, j;
    int prev_c, c;
    int len;
    spBool quoted = SP_FALSE;
    spBool have_space = SP_FALSE;

    if (input == NULL || (fp == NULL && line == NULL)) {
        return -1;
    }

    len = (int)strlen(input);
    if (len <= 0) {
        have_space = SP_TRUE;
    } else {
        if (len >= 2
            && ((input[0] == '\'' && input[len - 1] == '\'')
                || (input[0] == '\"' && input[len - 1] == '\"'))) {
            quoted = SP_TRUE;
        }
        
        for (i = 0; i < len; i++) {
            if (input[i] == ' ' || input[i] == '\t') {
                have_space = SP_TRUE;
                break;
            }
        }
    }

    j = 0;
    if (quoted == SP_FALSE && have_space == SP_TRUE) {
        if (fp != NULL) {
            putc('\"', fp);
        } else {
            if (size < 3) {
                line[0] = NUL;
                return 0;
            }
            line[j] = '\"';
        }
        j++;
    }

    prev_c = 0;
    for (i = 0;; i++) {
        c = input[i];
        if (c == NUL) {
            break;
        } else if (c == '#'
                   || (c == '\\'
                       && spIsMBTailCandidate(prev_c, c) == SP_FALSE)
                   || c == '\'' || c == '\"') {
            spDebug(80, "_spPutString", "%c found: prev_c = %c\n", c, prev_c);
            if (fp != NULL) {
                putc('\\', fp);
                putc(c, fp);
            } else {
                if (j + 2 >= size) {
                    break;
                }
                line[j] = '\\';
                line[j+1] = c;
            }
            j += 2;
        } else {
            if (fp != NULL) {
                putc(c, fp);
            } else {
                if (j + 1 >= size) {
                    break;
                }
                line[j] = c;
            }
            if (spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
                c = NUL;
            }
            
            j++;
        }
        prev_c = c;
    }

    if (quoted == SP_FALSE && have_space == SP_TRUE) {
        if (fp != NULL) {
            putc('\"', fp);
        } else {
            if (j + 2 >= size) {
                j = size - 2;
            }
            line[j] = '\"';
        }
        j++;
    }
    if (fp == NULL) {
        line[j] = NUL;
    }

    return j;
}

int spFPutString(const char *input, FILE *fp)
{
    return _spPutString(input, NULL, 0, fp);
}

int spSPutString(const char *input, char *line, int size)
{
    return _spPutString(input, line, size, NULL);
}

char *spFGetNString(char *buf, int size, FILE *fp)
{
    int c, prev_c;
    int pos;

    if (fp == NULL) return NULL;
    
    pos = 0;
    c = NUL;
    prev_c = NUL;
    while (1) {
        if (pos >= size) {
            spDebug(80, "spFGetNString", "exceed size: pos = %d / %d\n", pos, size);
            pos--;
            break;
        }

        c = fgetc(fp);
        if (c == EOF) {
            buf[pos] = NUL;
            if (pos <= 0) {
                spDebug(80, "spFGetNString", "no data: pos = %d / %d\n", pos, size);
                return NULL;
            } else {
                spDebug(80, "spFGetNString", "EOF found: pos = %d / %d\n", pos, size);
                return buf;
            }
        } else if (c == '\n' || c == '\r') {
            if (c == '\r' && (c = fgetc(fp)) != '\n') { /* for Mac */
                ungetc(c, fp);
            }
            if (prev_c == '\\') {
                pos--;
            } else {
                spDebug(80, "spFGetNString", "new line found: pos = %d / %d\n", pos, size);
                break;
            }
        } else {
            buf[pos] = c;
            pos++;

            if (/*c == '\\' &&*/ spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
                c = NUL;
            }
        }
        prev_c = c;
    }
    buf[pos] = NUL;

    spDebug(100, "spFGetNString", "done: pos = %d / %d, buf = %s\n", pos, size, buf);
    
    return buf;
}

char *spGetNString(char *buf, int size)
{
    return spFGetNString(buf, size, spgetstdin());
}

char *spSGetNString(char *buf, int size, const char *string)
{
    int c, prev_c;
    int pos;

    if (string == NULL || *string == NUL) return NULL;
    
    pos = 0;
    c = NUL;
    prev_c = NUL;
    while (1) {
        if (pos >= size) {
            pos--;
            break;
        }

        c = *string++;
        if (c == NUL) {
            buf[pos] = NUL;
            if (pos <= 0) {
                return NULL;
            } else {
                return (char *)(string - 1);
            }
        } else if (c == '\n' || c == '\r') {
            if (c == '\r' && (c = *string++) != '\n') { /* for Mac */
                --string;
            }
            if (prev_c == '\\') {
                pos--;
            } else {
                break;
            }
        } else {
            buf[pos] = c;
            pos++;

            if (/*c == '\\' &&*/ spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
                c = NUL;
            }
        }
        prev_c = c;
    }
    buf[pos] = NUL;

    return (char *)string;
}

int spSScanNSetup(const char *line, char *name, int namelen, char *value, int valuelen)
{
    int i;
    int end;
    int ncol;
    int nchar;
    int c, prev_c;

    name[0] = NUL;
    value[0] = NUL;
    
    prev_c = NUL;
    end = 0;
    ncol = 0;
    nchar = 0;
    for (i = 0;; i++) {
        c = line[i];
        /*spDebug(100, "spSScanNSetup", "c = %c\n", c);*/
        if (c == NUL) {
            c = EOF;
            break;
        } else if (c == '#' && prev_c != '\\') {
            for (i = 0;; i++) {
                c = line[i];
                if (c == NUL) {
                    end = 1;
                    break;
                } else if (c == '\n' || c == '\r') {
                    if (c == '\r' && (c = line[++i]) != '\n') { /* for Mac */
                        --i;
                    }
                    if (ncol > 0) {
                        end = 1;
                    }
                    break;
                }
            }
            
            if (end == 1) {
                c = EOF;
                break;
            }
        } else if (c == '\n' || c == '\r') {
            if (c == '\r' && (c = line[++i]) != '\n') { /* for Mac */
                --i;
            }
            if (prev_c == '\\' || prev_c == NUL) {
            } else if (ncol != 0 || nchar != 0) {
                c = EOF;
                break;
            }
        } else if ((c == ' ' || c == '\t') && prev_c != '\\') {
            if (nchar > 0 && ncol == 0) {
                if (nchar >= namelen) {
                    if (nchar == namelen) {
                        name[nchar - 1] = NUL;
                    }
                } else {
                    name[nchar] = NUL;
                }
                nchar = 0;
                ncol++;
            } else if (ncol != 0 && !(prev_c == ' ' || prev_c == '\t')) {
                if (nchar >= valuelen) {
                    if (nchar == valuelen) {
                        value[nchar - 1] = NUL;
                    }
                } else {
                    value[nchar] = ' ';
                }
                nchar++;
            }
        } else if (c == '\\' && prev_c != '\\'
                   && spIsMBTailCandidate(prev_c, c) == SP_FALSE) {
        } else {
            if (c == '\t') {
                c = ' ';
            }
            
            if (ncol == 0) {
                if (nchar >= namelen) {
                    if (nchar == namelen) {
                        name[nchar - 1] = NUL;
                    }
                } else {
                    name[nchar] = c;
                }
            } else {
                if (nchar >= valuelen) {
                    if (nchar == valuelen) {
                        value[nchar - 1] = NUL;
                    }
                } else {
                    if (prev_c == '\\' && c == 'n') {
                        value[nchar] = '\n';
                    } else {
                        value[nchar] = c;
                    }
                }
            }
            nchar++;

            if (c == '\\') {
                c = NUL;
            }
        }

        if (spIsMBTailCandidate(prev_c, c) == SP_TRUE) {
            prev_c = 0;
        } else {
            prev_c = c;
        }
    }

    if (nchar > 0) {
        if (ncol == 0) {
            nchar = MIN(nchar, namelen - 1);
        
            name[nchar] = NUL;
        } else {
            nchar = MIN(nchar, valuelen - 1);
            
            for (i = nchar - 1; i >= 0; i--) {
                if (value[i] == ' ' || value[i] == '\t') {
                    --nchar;
                } else {
                    break;
                }
            }

            if ((nchar >= 1 && value[nchar - 1] == '\\') ||
                nchar == 0) {
                ++nchar;
            }
            value[nchar] = NUL;
        }
        ncol++;
    }

    if (ncol <= 0 && c == EOF) {
        ncol = EOF;
    }

    return ncol;
}

int spSScanSetup(const char *line, char *name, char *value)
{
    return spSScanNSetup(line, name, SP_MAX_SETUP_NAME, value, SP_MAX_SETUP_VALUE);
}

/* the following is from g711.c */
/*
 * This source code is a product of Sun Microsystems, Inc. and is provided
 * for unrestricted use.  Users may copy or modify this source code without
 * charge.
 *
 * SUN SOURCE CODE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING
 * THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * Sun source code is provided with no support and without any obligation on
 * the part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS SOFTWARE
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

/*
 * g711.c
 *
 * u-law, A-law and linear PCM conversions.
 */
#define        SIGN_BIT        (0x80)                /* Sign bit for a A-law byte. */
#define        QUANT_MASK        (0xf)                /* Quantization field mask. */
#define        NSEGS                (8)                /* Number of A-law segments. */
#define        SEG_SHIFT        (4)                /* Left shift for segment number. */
#define        SEG_MASK        (0x70)                /* Segment field mask. */

/* copy from CCITT G.711 specifications */
static unsigned char _u2a[128] = {                        /* u- to A-law conversions */
        1,        1,        2,        2,        3,        3,        4,        4,
        5,        5,        6,        6,        7,        7,        8,        8,
        9,        10,        11,        12,        13,        14,        15,        16,
        17,        18,        19,        20,        21,        22,        23,        24,
        25,        27,        29,        31,        33,        34,        35,        36,
        37,        38,        39,        40,        41,        42,        43,        44,
        46,        48,        49,        50,        51,        52,        53,        54,
        55,        56,        57,        58,        59,        60,        61,        62,
        64,        65,        66,        67,        68,        69,        70,        71,
        72,        73,        74,        75,        76,        77,        78,        79,
        81,        82,        83,        84,        85,        86,        87,        88,
        89,        90,        91,        92,        93,        94,        95,        96,
        97,        98,        99,        100,        101,        102,        103,        104,
        105,        106,        107,        108,        109,        110,        111,        112,
        113,        114,        115,        116,        117,        118,        119,        120,
        121,        122,        123,        124,        125,        126,        127,        128};

static unsigned char _a2u[128] = {                        /* A- to u-law conversions */
        1,        3,        5,        7,        9,        11,        13,        15,
        16,        17,        18,        19,        20,        21,        22,        23,
        24,        25,        26,        27,        28,        29,        30,        31,
        32,        32,        33,        33,        34,        34,        35,        35,
        36,        37,        38,        39,        40,        41,        42,        43,
        44,        45,        46,        47,        48,        48,        49,        49,
        50,        51,        52,        53,        54,        55,        56,        57,
        58,        59,        60,        61,        62,        63,        64,        64,
        65,        66,        67,        68,        69,        70,        71,        72,
        73,        74,        75,        76,        77,        78,        79,        79,
        80,        81,        82,        83,        84,        85,        86,        87,
        88,        89,        90,        91,        92,        93,        94,        95,
        96,        97,        98,        99,        100,        101,        102,        103,
        104,        105,        106,        107,        108,        109,        110,        111,
        112,        113,        114,        115,        116,        117,        118,        119,
        120,        121,        122,        123,        124,        125,        126,        127};

static short seg_end[8] = {0xFF, 0x1FF, 0x3FF, 0x7FF,
                            0xFFF, 0x1FFF, 0x3FFF, 0x7FFF};

static int search(int val, short *table, int size)
{
        int                i;

        for (i = 0; i < size; i++) {
                if (val <= *table++)
                        return (i);
        }
        return (size);
}

/*
 * linear2alaw() - Convert a 16-bit linear PCM value to 8-bit A-law
 *
 * linear2alaw() accepts an 16-bit integer and encodes it as A-law data.
 *
 *                Linear Input Code        Compressed Code
 *        ------------------------        ---------------
 *        0000000wxyza                        000wxyz
 *        0000001wxyza                        001wxyz
 *        000001wxyzab                        010wxyz
 *        00001wxyzabc                        011wxyz
 *        0001wxyzabcd                        100wxyz
 *        001wxyzabcde                        101wxyz
 *        01wxyzabcdef                        110wxyz
 *        1wxyzabcdefg                        111wxyz
 *
 * For further information see John C. Bellamy's Digital Telephony, 1982,
 * John Wiley & Sons, pps 98-111 and 472-476.
 */
unsigned char spLinear2ALaw(
                          int pcm_val)        /* 2's complement (16-bit range) */
{
        int                mask;
        int                seg;
        unsigned char        aval;

        if (pcm_val >= 0) {
                mask = 0xD5;                /* sign (7th) bit = 1 */
        } else {
                mask = 0x55;                /* sign bit = 0 */
                pcm_val = -pcm_val - 8;
        }

        /* Convert the scaled magnitude to segment number. */
        seg = search(pcm_val, seg_end, 8);

        /* Combine the sign, segment, and quantization bits. */

        if (seg >= 8)                /* out of range, return maximum value. */
                return (0x7F ^ mask);
        else {
                aval = seg << SEG_SHIFT;
                if (seg < 2)
                        aval |= (pcm_val >> 4) & QUANT_MASK;
                else
                        aval |= (pcm_val >> (seg + 3)) & QUANT_MASK;
                return (aval ^ mask);
        }
}

/*
 * alaw2linear() - Convert an A-law value to 16-bit linear PCM
 *
 */
int spALaw2Linear(unsigned char a_val)
{
        int                t;
        int                seg;

        a_val ^= 0x55;

        t = (a_val & QUANT_MASK) << 4;
        seg = ((unsigned)a_val & SEG_MASK) >> SEG_SHIFT;
        switch (seg) {
        case 0:
                t += 8;
                break;
        case 1:
                t += 0x108;
                break;
        default:
                t += 0x108;
                t <<= seg - 1;
        }
        return ((a_val & SIGN_BIT) ? t : -t);
}

#define        BIAS                (0x84)                /* Bias for linear code. */

/*
 * linear2ulaw() - Convert a linear PCM value to u-law
 *
 * In order to simplify the encoding process, the original linear magnitude
 * is biased by adding 33 which shifts the encoding range from (0 - 8158) to
 * (33 - 8191). The result can be seen in the following encoding table:
 *
 *        Biased Linear Input Code        Compressed Code
 *        ------------------------        ---------------
 *        00000001wxyza                        000wxyz
 *        0000001wxyzab                        001wxyz
 *        000001wxyzabc                        010wxyz
 *        00001wxyzabcd                        011wxyz
 *        0001wxyzabcde                        100wxyz
 *        001wxyzabcdef                        101wxyz
 *        01wxyzabcdefg                        110wxyz
 *        1wxyzabcdefgh                        111wxyz
 *
 * Each biased linear code has a leading 1 which identifies the segment
 * number. The value of the segment number is equal to 7 minus the number
 * of leading 0's. The quantization interval is directly available as the
 * four bits wxyz.  * The trailing bits (a - h) are ignored.
 *
 * Ordinarily the complement of the resulting code word is used for
 * transmission, and so the code word is complemented before it is returned.
 *
 * For further information see John C. Bellamy's Digital Telephony, 1982,
 * John Wiley & Sons, pps 98-111 and 472-476.
 */
unsigned char spLinear2ULaw(
                          int pcm_val)        /* 2's complement (16-bit range) */
{
        int                mask;
        int                seg;
        unsigned char        uval;

        /* Get the sign and the magnitude of the value. */
        if (pcm_val < 0) {
                pcm_val = BIAS - pcm_val;
                mask = 0x7F;
        } else {
                pcm_val += BIAS;
                mask = 0xFF;
        }

        /* Convert the scaled magnitude to segment number. */
        seg = search(pcm_val, seg_end, 8);

        /*
         * Combine the sign, segment, quantization bits;
         * and complement the code word.
         */
        if (seg >= 8)                /* out of range, return maximum value. */
                return (0x7F ^ mask);
        else {
                uval = (seg << 4) | ((pcm_val >> (seg + 3)) & 0xF);
                return (uval ^ mask);
        }

}

/*
 * ulaw2linear() - Convert a u-law value to 16-bit linear PCM
 *
 * First, a biased linear code is derived from the code word. An unbiased
 * output can then be obtained by subtracting 33 from the biased code.
 *
 * Note that this function expects to be passed the complement of the
 * original code word. This is in keeping with ISDN conventions.
 */
int spULaw2Linear(
                unsigned char u_val)
{
        int                t;

        /* Complement to obtain normal u-law value. */
        u_val = ~u_val;

        /*
         * Extract and bias the quantization bits. Then
         * shift up by the segment number and subtract out the bias.
         */
        t = ((u_val & QUANT_MASK) << 3) + BIAS;
        t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;

        return ((u_val & SIGN_BIT) ? (BIAS - t) : (t - BIAS));
}

/* A-law to u-law conversion */
unsigned char spALaw2ULaw(
                        unsigned char aval)
{
        aval &= 0xff;
        return ((aval & 0x80) ? (0xFF ^ _a2u[aval ^ 0xD5]) :
            (0x7F ^ _a2u[aval ^ 0x55]));
}

/* u-law to A-law conversion */
unsigned char spULaw2ALaw(
                        unsigned char uval)
{
        uval &= 0xff;
        return ((uval & 0x80) ? (0xD5 ^ (_u2a[0xFF ^ uval] - 1)) :
            (0x55 ^ (_u2a[0x7F ^ uval] - 1)));
}
